/*
 * thread.h
 * Thread support.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#ifndef __thread_h
#define __thread_h

#include "md.h"

/*
 * Use native threads is available, otherwise use the internal ones.
 */
#if defined(USE_NATIVE_THREADS)
#include "thread-native.h"
#else
#include "thread-internal.h"
#endif

#define	THREADCLASS			"java/lang/Thread"
#define	THREADGROUPCLASS		"java/lang/ThreadGroup"
#define	THREADDEATHCLASS		"java/lang/ThreadDeath"

#define	MIN_THREAD_PRIO			1
#define	NORM_THREAD_PRIO		5
#define	MAX_THREAD_PRIO			10

#define	NOTIMEOUT			0

struct Hjava_lang_Object;

/* This structure mirrors java.lang.ThreadGroup.h */
typedef struct Hjava_lang_ThreadGroup {
	struct Hjava_lang_Object	obj;
	struct Hjava_lang_ThreadGroup*	parent;
	struct Hjava_lang_Object*	name;
	jint				maxPrio;
	jint				destroyed;
	jint				daemon;
	jint				vmAllowSuspension;
	jint				nthreads;
	struct Hjava_lang_Object*	threads;
	jint				ngroups;
	struct Hjava_lang_Object*	groups;
} Hjava_lang_ThreadGroup;

/* This structure mirrors java.lang.Thread.h */
typedef struct Hjava_lang_Thread {
	struct Hjava_lang_Object	obj;
	struct Hjava_lang_Object*	name;
	jint				priority;
	struct Hjava_lang_Thread*	next;
	jint				PrivateInfo;
	jint				eetop;
	jint				single_step;
	jint				daemon;
	jint				stillborn;
	struct Hjava_lang_Object*	target;
	struct Hjava_lang_ThreadGroup*	group;
	jint				initialStackMemory;
} Hjava_lang_Thread;

/*
 * Interface to the thread system.
 */
void	initThreads(void);
void	yieldThread(void);
void	sleepThread(jlong);
bool	aliveThread(Hjava_lang_Thread*);
jint	framesThread(Hjava_lang_Thread*);
void	setPriorityThread(Hjava_lang_Thread*, jint);
void	resumeThread(Hjava_lang_Thread*);
void	suspendThread(Hjava_lang_Thread*);
void	startThread(Hjava_lang_Thread*);
void	stopThread(Hjava_lang_Thread*);
extern int threadStackSize;
extern Hjava_lang_Thread* currentThread;

#endif
